# Copyright (C) 2008 Red Hat, Inc

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# a long with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

$: << File.dirname(File.expand_path(__FILE__))

gem 'reststop', '~> 0.2'

require 'yaml'
require 'fileutils'
require 'rubygems'
require 'picnic'
require 'reststop'
require 'genome-dsl'
require 'genome-lib'
require 'restr'
require 'cgi'
require 'xmlrpc/client'
require 'pp'

#BUG: the uri_path is not supported by reststop.  We should see if a newer
#version fixes this.  The problem is that the regex parsing in reststop is not
#handling contexts.  All this silly stuff with CONTEXT and our httpd
#proxypassing is a workaround. Yay.
#$CONF[:uri_path] ||= "genome"
#$CONF[:context] ||= "/genome"
#$CONF[:base_node_dir] ||= "/etc/genome/nodes"
#$CONF[:default_cloudmaster] ||= "DEAFULT CLOUDMASTER"

# We have to provide default values to be backwards compatible.  RPMs should
# overwrite user configuration.
#CONTEXT = Picnic::Conf[:context]
#BASE_NODE_DIR = Picnic::Conf[:base_node_dir]
#DEFAULT_CLOUDMASTER = Picnic::Conf.default_cloudmaster

Camping.goes :Genomed
Genomed.picnic!

module Genomed::Helpers
  include GenomeLib::Appliances

  BASE_NODE_DIR = Picnic::Conf.base_node_dir
  DEFAULT_CLOUDMASTER = Picnic::Conf.default_cloudmaster
  CONTEXT = Picnic::Conf.context
  puts "CONTEXT: #{CONTEXT}"

  def persist_node(node, yaml)
    node_dir = File.join(BASE_NODE_DIR, node)
    FileUtils.mkdir_p(node_dir) unless File.exist?(node_dir)
    File.open(node_dir + "/node.yaml", "w") {|f| f.puts yaml}
  end

  def hostname
    @hostname ||= `hostname`.chomp
  end

  def cloudmaster(fqdn)
    @cloudmaster ||= Cloud.new(fqdn, hostname)
  end

  def repo
    @repo ||= Genome.new(hostname)
  end
end


module Genomed::Controllers
  class MachineTypes < REST 'machine_types'
    include GenomeDsl
    include Genomed::Helpers

    # GET /types/foo
    # GET /types/foo.xml
    def read(type)
      @machine = machines.find {|m| m.name == type}
      render :view_type
    end

    # GET /types
    # GET /types.xml
    def list
      @machines = machines
      render :list_types
    end
  end

  class Nodes < REST 'nodes'
    include Genomed::Helpers

    # GET /nodes/foo
    # GET /nodes/foo.xml
    def read(name)
      @name = name
      begin
        @info = YAML.load(File.read("/etc/genome/nodes/#{@name}/node.yaml")) 
        render :view_node
      rescue
        # node doesn't exist yet
        render :add
      end
    end

    # GET /nodes
    # GET /nodes.xml
    def list
      @nodes = Dir["/etc/genome/nodes/*"].map { |m| File.basename(m) }
      render :list_nodes
    end

    def create
      if input.wizard
        puts "You're using the wizard!"   
        puts input.inspect
      elsif input.yaml
        node_dir = File.join(BASE_NODE_DIR, input.node_name)
        persist_node(input.node_name, input.yaml)
      elsif input.xml
        xml = XmlSimple.xml_in(input.xml)

        params = Hash.new
        xml["parameters"].each do |p| 
          p.each_key do |k|
            params[k] = p[k].to_s #The value is an array. Yay.
          end
        end

        persist_node(xml["name"].to_s, YAML.dump({"classes" => xml["class"], "parameters" => params}))
      end

      redirect R(Nodes) if @format == :HTML 
    end

    def destroy(node)
      node_dir = File.join(BASE_NODE_DIR, node)
      FileUtils.rm_rf(node_dir)
      $LOG.info(node_dir + " deleted")

      # TODO: When we tighten down the security on genomed this will need to be
      # fixed
      `/usr/sbin/puppetca --clean #{node}` 
      $LOG.info("Removed puppet cert for #{node}")
    end
  end

  class UpdateFQDN < REST 'update_fqdn'
    include Genomed::Helpers

    # POST /update_fqdn 
    def create
      @old_dir = "/etc/genome/nodes/#{input.old_fqdn}"
      @new_dir = "/etc/genome/nodes/#{input.new_fqdn}"
      if File.exist?(@old_dir) and File.stat(@old_dir).directory?
        File.rename(@old_dir, @new_dir)
        machine_config = YAML.load("#{@new_dir}/node.yaml")
        cloud = cloudmaster(machine_config["parameters"]["cloudmaster"])
        cloud.update_nodename(input.old_fqdn, input.new_fqdn)
        @status = 200
      else
        @status = 500
      end
      render :view_update_fqdn
    end
  end

  class Wizard < REST 'wizard'
    include GenomeDsl
    include Genomed::Helpers

    # GET /wizard/foo
    # GET /wizard/foo.xml
    def read(type)
      @machine = machines.find {|m| m.name == type}
      render :view_type
    end

    # GET /types
    # GET /types.xml
    def list
      @machines = machines
      render :list_types
    end

    def create
      @machine_type = input.machine_type

      redirect CONTEXT + R(Nodes)
    end
  end

  class Bootstrap < REST 'bootstrap'
    include GenomeDsl
    include Genomed::Helpers


    # GET /bootstrap.html
    # The "first page" for bootstrapping
    def list 
      render :bootstrap_start 
    end

    # POST /bootstrap
    # Handle a multi-step "wizard" process for bootstrapping machines.
    # Each POST determines which step to execute.
    def create
      if input.config
        @config = YAML.load(input.config)
        if input.parameters
          # merge in the incoming parameters
          @config["parameters"].merge!(input.parameters)
        end
      else
        @config ||= {"parameters" => { "repoappliance" => hostname }, "classes" => []}
      end

      if input.next_step
        next_step = validate ? input.next_step : input.this_step
        begin
          method(next_step).to_proc.call
        rescue
          render "bootstrap_#{next_step}".to_sym
        end
      end
    end

    private 

    def validate
      @errors = {}
      case input.this_step.to_sym
      when :choose_machine_type
        if @config["parameters"]["genome_machine_type"] == "Select one"
          @errors["parameters[genome_machine_type]"] = "empty"
        end
        if @config["parameters"]["machine_prefix"] == ""
          @errors["parameters[machine_prefix]"] = "empty"
        end
      when :collect_data
        # no validation
      when :select_profile
        if @config["parameters"]["cobbler_profile"] == "Select one"
          @errors["parameters[cobbler_profile]"] = "empty"
        end
      when :select_cloudmaster
        if @config["parameters"]["cloudmaster"] == ""
          @errors["parameters[cloudmaster]"] = "empty"
        else
          begin
            cloud = cloudmaster(@config["parameters"]["cloudmaster"])
            largest_machine = cloud.status("memory")[0]["largest_machine"].to_i
            cobbler_profile = repo.cobbler_call("get_profiles").select {|p| 
              p["name"] == @config["parameters"]["cobbler_profile"] 
            }[0]
            virt_ram = cobbler_profile["virt_ram"].to_i
            puts "largest_machine: #{largest_machine}, virt_ram: #{virt_ram}"
            if largest_machine < virt_ram
              @errors["parameters[cloudmaster]"] = "profile_too_large"
            end
          rescue GenomeLib::Appliances::InvalidAppliance
            @errors["parameters[cloudmaster]"] = "invalid"
          end
        end
        if @config["parameters"]["contact_email"] == ""
          @errors["parameters[contact_email]"] = "empty"
        end
      when :provision
        # no validation
      when :review
        # no validation
      end

      return @errors.empty?
    end
    
    def choose_machine_type
      @machines = machines
      render :bootstrap_choose_machine_type
    end

    def collect_data
      @machine_type = @config["parameters"]["genome_machine_type"]
      @machine = machines.find {|m| m.name == @machine_type}
      @config["parameters"]["genome_system_name"] = 
          "#{@config["parameters"]["machine_prefix"]}-#{@config["parameters"]["genome_machine_type"]}"

      @config["classes"] = @machine.classes
      render :bootstrap_collect_data
    end

    def complete
      system_name = @config["parameters"]["genome_system_name"]
      email_addr = @config["parameters"]["contact_email"]
      cloud_host = @config["parameters"]["cloudmaster"]
      repo.register_machine(system_name, @config)
      id, @host = cloudmaster(cloud_host).create_machine(system_name, email_addr)
      render :bootstrap_complete
    end
  end
end

module Genomed::Views
  def initialize(*args)
    super(*args)
    @x = Builder::XmlMarkup.new(:indent => 1)
  end

  module HTML
    include Genomed::Helpers

    def layout
      html do
        head do
          title 'Genome'
          link :rel => 'stylesheet', :type => 'text/css', :href => '/genome-styling/styles.css', :media => 'screen'
        end
        body do
          div.header! do
            div.headerLogo! {}
          end
          div.body! do
            div.nav_menu! do
              ul do
                puts "route to machine types: #{R(Genomed::Controllers::MachineTypes)}"
                puts "route to machine types (2): #{CONTEXT + R(Genomed::Controllers::MachineTypes)}"
                puts "CONTEXT: #{CONTEXT}"
                li { a("Machine Types", :href => CONTEXT + R(Genomed::Controllers::MachineTypes)) } 
                li { a("Nodes", :href => CONTEXT + R(Genomed::Controllers::Nodes)) }
                li { a("Bootstrap", :href => CONTEXT + R(Genomed::Controllers::Bootstrap)) }
              end
            end
            div.content! do
              div.inner! do
                self << yield
              end
            end
          end
        end
      end
    end
 
    def add
      form(:method => 'post', :action => CONTEXT + R(Genomed::Controllers::Nodes)) do
        h2 "YAML"
 
        textarea :name => 'yaml', :rows => '25', :cols => '100'; br

        input :name => 'node_name', :value => @name
        input :type => 'submit', :value => "Submit"
      end
    end

    def view_type
      p do
        h1 @machine.name
        p @machine.desc

        form(:method => 'post', :action => CONTEXT + R(Genomed::Controllers::Nodes)) do
          h2 "Fully qualified domain name"
          input :name => "fqdn", :type => 'text', :size  => 45,
                :value => "[prefix]-#{@machine.name}.usersys.redhat.com"

          h2 "Classes"
          @machine.classes.each_with_index do |c, i|
            input :name => "class#{i}", :type => 'text', 
                  :value => c; br
          end
  
          h2 "Parameters"
          fieldset do
            @machine.facts.each do |f|
              a(f.name, :href => "fix me", :title => f.desc); br
              input :name => "param_" + f.name, :type => 'text', :size => 45, 
                    :value => f.default; br
            end
          end

          input :type => 'hidden', :name => 'node_name', :value => @name
          input :type => 'hidden', :name => 'wizard', :value => true
          input :type => 'submit', :value => "Submit"
        end
      end
    end 

    def view_node
      h1 @name
    
      p "This form allows you to submit the YAML to be used by the Puppetmaster to " +
        "determine the configuration to be compiled for a particular host."
      form(:method => 'post', :action => CONTEXT + R(Genomed::Controllers::Nodes)) do
 
        textarea ::YAML.dump(@info), :name => 'yaml', :rows => '25', :cols => '100'; br

        input :type => 'hidden', :name => 'node_name', :value => @name
        input :type => 'submit', :value => "Submit"
      end
    end

    def view_update_fqdn
      if @status == 200
        div do
          h1 "FQDN Updated"
          h3 "old: #{@old_dir}"
          h3 "new: #{@new_dir}"
        end
      else
        div do
          h1 "FQDN NOT Updated"
          h3 "directory #{@old_dir} does not exist"
        end
      end
    end


    def list_nodes
      h1 "Nodes:"
      p "To create a new node just browse to the page corresponding to the hostname.  " +
        "eg, #{CONTEXT + R(Genomed::Controllers::Nodes, "[hostname]")}."
      p do
        ol do
          @nodes.each do |n| 
            # BUG: default format is not getting preserved
            @format = :HTML
            li {a(n, :href => CONTEXT + R(Genomed::Controllers::Nodes, n))}
          end
        end
      end
    end

    def list_types
      p do
        p "Click the links to create a new instance of a particular machine type."
        ol do
          @machines.each do |m|
            li do 
              #h1 {a(m.name, :href => R(Genomed::Controllers::MachineTypes, m.name))}
              h1 m.name
              p(m.desc)
            end
          end
        end
      end
    end

    def bootstrap_start
      p do
        h2 "Genome Bootstrap"
        h3 "Beta", :class => "warning" 
        p "This is the Genome Bootstrap web tool.  This is a beta release " +
          "so if you're not ready for a beta tool, stick to the command line genome-bootstrap."
        p "Ready for beta? "
        p do
          form(:method => "post", :action => CONTEXT + R(Genomed::Controllers::Bootstrap)) do
            input :type => "hidden", :name => "next_step", :value => "choose_machine_type"
            input :type => "hidden", :name => "this_step", :value => "start"
            input :type => "submit", :value => "Select machine type"
          end
        end
      end
    end

    def bootstrap_choose_machine_type
      p do
        h2 "Genome Bootstrap"
        h3 "Step 1:  Select machine type"

        form(:method => 'post', :action => CONTEXT + R(Genomed::Controllers::Bootstrap)) do

          input :type => "hidden", :name => "next_step", :value => "collect_data"
          input :type => "hidden", :name => "this_step", :value => "choose_machine_type"
          input :type => "hidden", :name => "config", :value => "#{::YAML.dump(@config)}"

          fieldset do
            table do
              tr do
                td.label do text("Machine type:") end
                td.field do  
                  tag!(:select, :id => "machine_type", :name => "parameters[genome_machine_type]") do
                    tag!(:option, "Select one")
                    @machines.each do |machine|
                      if machine.name == @config["parameters"]["genome_machine_type"]
                        tag!(:option, "#{machine.name}", :selected => "selected")
                      else
                        tag!(:option, "#{machine.name}")
                      end
                    end
                  end
                end
                td(:class => "error", :style => (@errors.keys.include? "parameters[genome_machine_type]") ? "visibility:visible" : "visibility:hidden") do
                  text("Please select a machine type")
                end
              end
              tr do
                td.label do text("Machine prefix:") end
                td.field do
                  input :type => "text", :id => "machine_prefix", :name => "parameters[machine_prefix]", :value => @config["parameters"]["machine_prefix"]
                end
                td(:class => "error", :style => (@errors.keys.include? "parameters[machine_prefix]") ? "visibility:visible" : "visibility:hidden") do
                  text("Please enter a machine prefix")
                end
              end
            end
          end
          p do
            input :type => "submit", :value => "Next: Collect Data"
          end
        end
      end
    end

    def bootstrap_collect_data
      p do
        h2 "Genome Bootstrap"
        h3 "Step 2: Collect bootstrap data"
      end
      p do
        form(:method => "post", :action => CONTEXT + R(Genomed::Controllers::Bootstrap)) do
          input :type => "hidden", :name => "next_step", :value => "select_profile"
          input :type => "hidden", :name => "this_step", :value => "collect_data"
          input :type => "hidden", :name => "config", :value => "#{::YAML.dump(@config)}"
          
          fieldset do
            table do
              @machine.facts.each do |f|
                tr do
                  td.label do text("#{f.name.capitalize}:") end
                  td.field do
                    input :name => "parameters[#{f.name}]", :type => 'text', :size => 35,
                          :value => f.default.gsub("%repo%", `hostname`); br
                  end
                end
              end
            end
          end
          p do
            input :type => "submit", :value => "Next: Select profile"
          end
        end
      end
    end

    def bootstrap_select_profile
      p do
        h2 "Genome Bootstrap"
        h3 "Step 3: Select a machine profile"
      end
      p do
        form(:method => "post", :action => CONTEXT + R(Genomed::Controllers::Bootstrap)) do
          input :type => "hidden", :name => "next_step", :value => "select_cloudmaster"
          input :type => "hidden", :name => "this_step", :value => "select_profile"
          input :type => "hidden", :name => "config", :value => "#{::YAML.dump(@config)}"

          fieldset do
            table do
              tr do
                td.label do text("Machine profile") end
                td.field do
                  tag!(:select, :id => "cobbler_profile", :name => "parameters[cobbler_profile]") do
                    tag!(:option, "Select one")
                    repo.guest_profiles.each do |profile|
                      if profile == @config["parameters"]["cobbler_profile"]
                        tag!(:option, "#{profile}", :selected => "selected")
                      else
                        tag!(:option, "#{profile}")
                      end
                    end
                  end
                end
                td(:class => "error", :style => (@errors.keys.include? "parameters[cobbler_profile]") ? "visibility:visible" : "visibility:hidden") do
                  text("Please select a machine profile")
                end
              end
            end
          end

          p do
            input :type => "submit", :value => "Next: Select cloudmaster"
          end
        end
      end
    end

    def bootstrap_select_cloudmaster
      p do
        h2 "Genome Bootstrap"
        h3 "Step 4: Select the cloudmaster"
      end
      p do
        form(:method => "post", :action => CONTEXT + R(Genomed::Controllers::Bootstrap)) do
          input :type => "hidden", :name => "next_step", :value => "review"
          input :type => "hidden", :name => "this_step", :value => "select_cloudmaster"
          input :type => "hidden", :name => "config", :value => "#{::YAML.dump(@config)}"

          fieldset do
            table do
              tr do
                td.label do text("Cloudmaster domain name") end
                td.field do
                  params = @config["parameters"]
                  value = (params.has_key? "cloudmaster" and params["cloudmaster"] != "") ? params["cloudmaster"] : "#{Picnic::Conf.default_cloudmaster}" 
                  input :type => "text", :name => "parameters[cloudmaster]", :value => "", :size => "30", :value => value
                end
                td(:class => "error", :style => (@errors.keys.include? "parameters[cloudmaster]") ? "visibility:visible" : "visibility:hidden") do
                  if @errors["parameters[cloudmaster]"] == "empty"
                    text("Please enter a cloudmaster domain name")
                  elsif @errors["parameters[cloudmaster]"] == "invalid"
                    text("No cloudmaster running at that domain name")
                  elsif @errors["parameters[cloudmaster]"] == "profile_too_large"
                    text("This cloudmaster cannot hold your selected profile")
                  end
                end
              end
              tr do
                td.label do text("Your email address") end
                td.field do
                  input :type => "text", :name => "parameters[contact_email]", :value => @config["parameters"]["contact_email"], :size => "30"
                end
                td(:class => "error", :style => (@errors.keys.include? "parameters[contact_email]") ? "visibility:visible" : "visibility:hidden") do
                  text("Please enter you email address")
                end
              end
            end
          end

          p do
            input :type => "submit", :value => "Next: Review"
          end
        end
      end
    end

    def bootstrap_review
      p do
        h2 "Genome Bootstrap"
        h3 "Step 5: Review the final yaml"
      end
      p do
        form(:method => "post", :action => CONTEXT + R(Genomed::Controllers::Bootstrap)) do
          input :type => "hidden", :name => "next_step", :value => "complete"
          input :type => "hidden", :name => "this_step", :value => "review"
          
          textarea ::YAML.dump(@config), :name => 'config', :rows => '25', :cols => '100'; br

          input :type => "submit", :value => "Finish"
        end
      end
    end

    def bootstrap_complete
      p do
        h2 "Genome Bootstrap"
        h3 "Complete!"
      end
      p do
        p do
          text("Your virtual machine is now installing.  Note:  The installation process will take several minutes.")
          text("You can visit the cloudmaster to view the state of the virtual machine.")
          text("The cloud master will also be updated with the IP address of the virtual machine.")
        end
        p do
          text("Your machine is being provisioned on #{host}.")
        end
        p do
          a("Visit the cloudmaster", :href => "http://#{@config["parameters"]["cloudmaster"]}/cloud/status.html")
        end

      end
    end

    def bootstrap_invalid 
      h2 "Genome Bootstrap"
      h3 "Invalid step"
    end
        
  end
  default_format :HTML

  module YAML
    include Genomed::Helpers

    CONTENT_TYPE = 'text/plain'
    
    def layout
      yield
    end

    # This should only be used for Puppet's external nodes
    def view_node
      ::YAML.dump(@info)
    end
  end

  module XML
    include Genomed::Helpers

    def layout
      yield
    end

    def list_types
      @x.instruct!
      @x.machine_types {
        @machines.each do |m|
          @x.machine_type {
            @x.name(m.name)
            @x.description(m.desc)
          }
        end
      }
    end

    def view_type
      @x.instruct!
      @x.machine {
        @x.name(@machine.name)
        @x.description(@machine.desc)
        @machine.classes.each {|c| @x.class(c)}
        @machine.facts.each {|f| 
          @x.fact {
            @x.name(f.name)
            @x.description(f.desc)
            @x.default(f.default)
          }
        }
      }
    end

    def list_nodes
      @x.instruct!
      @x.nodes {
        @nodes.each {|n| @x.node(n)}
      }
    end

    def view_node
      # There no reason we should only support params and classes.
      # It might be nice to simply iterate over all the keys and 
      # display their values.
      @x.instruct!
      @x.node {
        @x.name(@name)
        @info["parameters"].each_pair do |n,v| 
          @x.parameters {
            instance_eval("@x.#{n}(v)")
          }
        end
        @info["classes"].each {|c| @x.class(c)}
      } 
    end
  end
end

# Required by picnic
def Genomed.create; end

Genomed.start_picnic
